Index: libs/libmythtv/NuppelVideoPlayer.cpp
===================================================================
RCS file: /var/lib/mythcvs/mythtv/libs/libmythtv/NuppelVideoPlayer.cpp,v
retrieving revision 1.312
diff -u -r1.312 NuppelVideoPlayer.cpp
--- libs/libmythtv/NuppelVideoPlayer.cpp	5 Jan 2004 02:09:04 -0000	1.312
+++ libs/libmythtv/NuppelVideoPlayer.cpp	9 Jan 2004 04:11:56 -0000
@@ -91,7 +91,8 @@
 
     nvr_enc = NULL;
 
-    paused = 0;
+    paused = false;
+    previously_paused = false;
 
     audiodevice = "/dev/dsp";
     audioOutput = NULL;
@@ -216,6 +217,9 @@
 
 void NuppelVideoPlayer::Pause(bool waitvideo)
 {
+    if (paused)
+        return;
+
     actuallypaused = false;
 
     // pause the decoder thread first, as it's usually waiting for the video
@@ -233,10 +237,32 @@
 
     if (ringBuffer)
         ringBuffer->Pause();
+
+    play_speed = 1.0;
+    normal_speed = false;
+    frame_interval = (int)(1000000.0 / video_frame_rate / 1.0);
+    if (osd && forceVideoOutput != kVideoOutput_IVTV)
+        osd->SetFrameInterval(frame_interval);
 }
 
-void NuppelVideoPlayer::Unpause(bool unpauseaudio)
+bool NuppelVideoPlayer::Play(float speed, bool normal, bool unpauseaudio)
 {
+    if (forceVideoOutput == kVideoOutput_IVTV && speed > 3.0)
+        return false;
+
+    play_speed = speed;
+    normal_speed = normal;
+    frame_interval = (int)(1000000.0 / video_frame_rate / speed);
+    if (osd && forceVideoOutput != kVideoOutput_IVTV)
+        osd->SetFrameInterval(frame_interval);
+
+    if (!paused)
+    {
+        // Force speed update for IVTV
+        previously_paused = true;
+        return true;
+    }
+
     paused = false;
     UnpauseVideo();
     if (audioOutput && unpauseaudio)
@@ -244,6 +270,8 @@
 
     if (ringBuffer)
         ringBuffer->Unpause();
+
+    return true;
 }
 
 bool NuppelVideoPlayer::GetPause(void)
@@ -276,15 +304,6 @@
     pausevideo = false;
 }
 
-void NuppelVideoPlayer::SetPlaySpeed(float speed, bool normal)
-{
-    play_speed = speed;
-    normal_speed = normal;
-    frame_interval = (int)(1000000.0 / video_frame_rate / speed);
-    if (osd)
-        osd->SetFrameInterval(frame_interval);
-}
-
 void NuppelVideoPlayer::setPrebuffering(bool prebuffer)
 {
     prebuffering_lock.lock();
@@ -1564,56 +1583,29 @@
 void NuppelVideoPlayer::IvtvVideoLoop(void)
 {
     refreshrate = frame_interval;
+    int delay = frame_interval;
 
     VideoOutputIvtv *vidout = (VideoOutputIvtv *)videoOutput;
-
     vidout->SetFPS(GetFrameRate());
 
-    bool previously_paused = false;
-
     while (!eof && !killvideo)
     {
+        resetvideo = false;
+        video_actually_paused = pausevideo;
+
         if (pausevideo)
         {
-            if (!previously_paused)
-                vidout->Pause();
-
-            if (!video_actually_paused)
-                videoOutput->UpdatePauseFrame();
-
-            if (resetvideo)
-            {
-                videoOutput->UpdatePauseFrame();
-                resetvideo = false;
-            }
-
-            video_actually_paused = true;
             videoThreadPaused.wakeAll();
-
             videoOutput->ProcessFrame(NULL, osd, videoFilters, pipplayer);
-
-            previously_paused = true;
-
-            //printf("video waiting for unpause\n");
-            usleep(frame_interval);
-            continue;
         }
-
-        if (previously_paused)
+        else
         {
-            vidout->Play();
-            previously_paused = false;
+            if (cc)
+                ShowText();
+            videoOutput->ProcessFrame(NULL, osd, videoFilters, NULL);
         }
 
-        video_actually_paused = false;
-        resetvideo = false;
-
-        if (cc)
-            ShowText();
-
-        videoOutput->ProcessFrame(NULL, osd, videoFilters, NULL);
-
-        usleep(frame_interval);
+        usleep(delay);
     }
 
     delete videoOutput;
@@ -1634,11 +1626,7 @@
 bool NuppelVideoPlayer::FastForward(float seconds)
 {
     if (fftime == 0)
-    {
         fftime = (int)(seconds * video_frame_rate);
-        if (decoder)
-            decoder->InterruptDisplay();
-    }
 
     return fftime > CalcMaxFFTime(fftime);
 }
@@ -1646,11 +1634,7 @@
 bool NuppelVideoPlayer::Rewind(float seconds)
 {
     if (rewindtime == 0)
-    {
         rewindtime = (int)(seconds * video_frame_rate);
-        if (decoder)
-            decoder->InterruptDisplay();
-    }
 
     return rewindtime >= framesPlayed;
 }
@@ -1756,6 +1740,17 @@
     {
         if (paused)
         { 
+            if (!previously_paused)
+            {
+                if (forceVideoOutput == kVideoOutput_IVTV)
+                {
+                    VideoOutputIvtv *vidout = (VideoOutputIvtv *)videoOutput;
+                    vidout->Pause();
+                }
+                decoder->UpdateFramesPlayed();
+                previously_paused = true;
+            }
+
             actuallypaused = true;
             decoderThreadPaused.wakeAll();
             pausecheck++;
@@ -1764,7 +1759,7 @@
             {
                 if (livetv && ringBuffer->GetFreeSpace() < -1000)
                 {
-                    Unpause();
+                    Play();
                     printf("forced unpause\n");
                 }
                 pausecheck = 0;
@@ -1803,6 +1798,16 @@
             }
         }
 
+        if (previously_paused)
+        {
+            if (forceVideoOutput == kVideoOutput_IVTV)
+            {
+                VideoOutputIvtv *vidout = (VideoOutputIvtv *)videoOutput;
+                vidout->Play(play_speed, normal_speed);
+            }
+            previously_paused = false;
+        }
+
         if (rewindtime > 0)
         {
             PauseVideo();
@@ -1856,7 +1861,7 @@
             if (deleteIter.key() == totalFrames)
                 eof = 1;
             else
-                fftime = deleteIter.key() - framesPlayed;
+                JumpToFrame(deleteIter.key());
             ++deleteIter;
         }
     }
@@ -2345,8 +2350,6 @@
    
     QMutexLocker lockit(&db_lock);
     m_playbackinfo->SetEditing(false, m_db);
-
-    Unpause();
 }
 
 bool NuppelVideoPlayer::DoKeypress(QKeyEvent *e)
@@ -3681,7 +3684,7 @@
         commBreakIter++;
 
         GetFrame(1, true);
-        while (LastFrameIsBlank())
+        while (LastFrameIsBlank() && !eof && !killplayer)
             GetFrame(1, true);
 
         return true;
Index: libs/libmythtv/NuppelVideoPlayer.h
===================================================================
RCS file: /var/lib/mythcvs/mythtv/libs/libmythtv/NuppelVideoPlayer.h,v
retrieving revision 1.132
diff -u -r1.132 NuppelVideoPlayer.h
--- libs/libmythtv/NuppelVideoPlayer.h	5 Jan 2004 02:09:04 -0000	1.132
+++ libs/libmythtv/NuppelVideoPlayer.h	9 Jan 2004 04:11:56 -0000
@@ -69,16 +69,11 @@
 
     void SetAudioSampleRate(int rate) { audio_samplerate = rate; }
 
-    bool TogglePause(void) { if (paused) Unpause(); else Pause(); 
-                             return paused; 
-                           }
     void Pause(bool waitvideo = true);
-    void Unpause(bool unpauseaudio = true); 
+    bool Play(float speed = 1.0, bool normal = true, 
+              bool unpauseaudio = true); 
     bool GetPause(void);
 
-    float GetPlaySpeed(void) { return play_speed; };
-    void SetPlaySpeed(float speed, bool normal);
-
     bool FastForward(float seconds);
     bool Rewind(float seconds);
 
@@ -295,7 +290,7 @@
 
     AudioOutput *audioOutput;
 
-    bool paused, pausevideo;
+    bool paused, previously_paused, pausevideo;
     bool actuallypaused, video_actually_paused;
     QWaitCondition decoderThreadPaused, videoThreadPaused;
  
Index: libs/libmythtv/decoderbase.h
===================================================================
RCS file: /var/lib/mythcvs/mythtv/libs/libmythtv/decoderbase.h,v
retrieving revision 1.11
diff -u -r1.11 decoderbase.h
--- libs/libmythtv/decoderbase.h	5 Jan 2004 02:09:04 -0000	1.11
+++ libs/libmythtv/decoderbase.h	9 Jan 2004 04:11:56 -0000
@@ -46,8 +46,8 @@
 
     virtual QString GetEncodingType(void) = 0;
 
-    virtual void InterruptDisplay(void) { }
-       
+    virtual void UpdateFramesPlayed(void) { };
+
   protected:
     NuppelVideoPlayer *m_parent;
     RingBuffer *ringBuffer;
Index: libs/libmythtv/ivtv-ext-api.h
===================================================================
RCS file: /var/lib/mythcvs/mythtv/libs/libmythtv/ivtv-ext-api.h,v
retrieving revision 1.7
diff -u -r1.7 ivtv-ext-api.h
--- libs/libmythtv/ivtv-ext-api.h	8 Nov 2003 23:59:26 -0000	1.7
+++ libs/libmythtv/ivtv-ext-api.h	9 Jan 2004 04:11:56 -0000
@@ -25,14 +25,23 @@
 #define IVTV_IOC_S_OSD          0xFFEE7788
 #define IVTV_IOC_GET_FB         0xFFEE7789
 
+#define IVTV_IOC_START_DECODE   _IOW('@', 29, struct ivtv_cfg_start_decode)
+#define IVTV_IOC_STOP_DECODE    _IOW('@', 30, struct ivtv_cfg_stop_decode)
+#define IVTV_IOC_G_SPEED        _IOR('@', 31, struct ivtv_speed)
+#define IVTV_IOC_S_SPEED        _IOW('@', 32, struct ivtv_speed)
+#define IVTV_IOC_DEC_STEP       _IOW('@', 33, int)
+#define IVTV_IOC_DEC_FLUSH      _IOW('@', 34, int)
+
+
 /* ioctl for MSP_SET_MATRIX will have to be registered */
 #define MSP_SET_MATRIX     _IOW('m',17,struct msp_matrix)
 
+
+/* Custom v4l controls */
 #ifndef V4L2_CID_PRIVATE_BASE
-#define V4L2_CID_PRIVATE_BASE               0x08000000
+#define V4L2_CID_PRIVATE_BASE			0x08000000
 #endif
 
-/* Custom v4l controls */
 #define V4L2_CID_IVTV_FREQ      (V4L2_CID_PRIVATE_BASE)
 #define V4L2_CID_IVTV_ENC       (V4L2_CID_PRIVATE_BASE + 1)
 #define V4L2_CID_IVTV_BITRATE   (V4L2_CID_PRIVATE_BASE + 2)
@@ -70,6 +79,18 @@
 	uint64_t scr;
 };
 
+struct ivtv_speed {
+	int scale;	/* 1-?? (50 for now) */
+	int smooth;	/* Smooth mode when in slow/fast mode */
+	int speed;	/* 0 = slow, 1 = fast */
+	int direction;	/* 0 = forward, 1 = reverse (not supportd */
+	int fr_mask;	/* 0 = I, 1 = I,P, 2 = I,P,B    2 = default!*/
+	int b_per_gop;	/* frames per GOP (reverse only) */
+	int aud_mute;	/* Mute audio while in slow/fast mode */
+	int fr_field;	/* 1 = show every field, 0 = show every frame */
+	int mute;	/* # of audio frames to mute on playback resume */
+};
+
 struct ivtv_slow_fast {
 	int speed; /* 0 = slow, 1 = fast */
 	int scale; /* 1-?? (50 for now) */
@@ -81,7 +102,8 @@
 };
 
 struct ivtv_cfg_stop_decode {
-	int		hide_last;/* 1 = show black after stop,0 = show last frame */
+	int		hide_last; /* 1 = show black after stop,
+				      0 = show last frame */
 	uint64_t	pts_stop; /* PTS to stop at */
 };
 
@@ -164,3 +186,6 @@
 #define IVTVFB_STATUS_GLOBAL_ALPHA      (1 << 1)
 #define IVTVFB_STATUS_LOCAL_ALPHA       (1 << 2)
 #define IVTVFB_STATUS_FLICKER_REDUCTION (1 << 3)
+
+#define IVTV_IOCTL_SET_DEBUG_LEVEL _IOWR('@', 98, int *)
+#define IVTV_IOCTL_GET_DEBUG_LEVEL _IOR('@', 99, int *)
Index: libs/libmythtv/ivtvdecoder.cpp
===================================================================
RCS file: /var/lib/mythcvs/mythtv/libs/libmythtv/ivtvdecoder.cpp,v
retrieving revision 1.17
diff -u -r1.17 ivtvdecoder.cpp
--- libs/libmythtv/ivtvdecoder.cpp	27 Dec 2003 21:10:07 -0000	1.17
+++ libs/libmythtv/ivtvdecoder.cpp	9 Jan 2004 04:11:56 -0000
@@ -27,15 +27,15 @@
     m_db = db;
     m_playbackinfo = pginfo;
 
-    framesPlayed = 0;
     framesRead = 0;
+    framesPlayed = 0;
+    lastStartFrame = 0;
 
     hasFullPositionMap = false;
 
     keyframedist = 15;
 
     exitafterdecoded = false;
-    ateof = false;
     gopset = false;
     firstgoppos = 0;
     prevgoppos = 0;
@@ -47,26 +47,54 @@
 
     prvpkt[0] = prvpkt[1] = prvpkt[2] = 0;
 
-    ingetframe = false;
+    vidread = vidwrite = vidfull = 0;
+    ateof = false;
 }
 
 IvtvDecoder::~IvtvDecoder()
 {
 }
 
-void IvtvDecoder::SeekReset(int skipframes)
+void IvtvDecoder::SeekReset(long long newkey, int skipframes)
 {
+    //fprintf(stderr, "seek reset frame = %llu, skip = %d, exact = %d\n", 
+    //        newkey, skipframes, exactseeks);
+
+    if (!exactseeks)
+        skipframes = 0;
+
+    vidread = vidwrite = vidfull = 0;
+    ateof = false;
+
+    framesRead = newkey;
+    lastStartFrame = newkey;
+
     VideoOutputIvtv *videoout = (VideoOutputIvtv *)m_parent->getVideoOutput();
-    videoout->Reopen(skipframes, framesPlayed);
+
+    videoout->Stop();
+    videoout->Flush();
+
+    videoout->Start(0, skipframes+5);
+    videoout->Pause();
+
+    while (ReadWrite(1, 0))
+        ;
+    while (!videoout->GetFramesPlayed() && !ateof)
+        ReadWrite(1, 0);
+
+    if (m_parent->GetPause())
+        StepFrames(newkey, skipframes);
+    else
+        videoout->Play();
+
+    framesPlayed = newkey + skipframes + 1;
+    m_parent->SetFramesPlayed(framesPlayed);
 }
 
 void IvtvDecoder::Reset(void)
 {
     positionMap.clear();
-    framesPlayed = 0;
-    framesRead = 0;
-
-    SeekReset();
+    SeekReset(0, 0);
 }
 
 bool IvtvDecoder::CanHandle(char testbuf[2048], const QString &filename)
@@ -84,16 +112,22 @@
 
     struct v4l2_capability vcap;
     memset(&vcap, 0, sizeof(vcap));
-    if (ioctl(testfd, VIDIOC_QUERYCAP, &vcap) >= 0)
-    {
-        if (vcap.capabilities & V4L2_CAP_VIDEO_OUTPUT)
+    if (ioctl(testfd, VIDIOC_QUERYCAP, &vcap) < 0)
+        perror("VIDIOC_QUERYCAP");
+    else
+    {
+        fprintf(stderr, "driver version = %x\n", vcap.version);
+        if (vcap.capabilities & V4L2_CAP_VIDEO_OUTPUT &&
+            vcap.version >= 0x00000108)
             ok = true;
     }
 
     v4l2_std_id std = V4L2_STD_NTSC;
     ntsc = true;
 
-    if (ioctl(testfd, VIDIOC_G_STD, &std) >= 0)
+    if (ioctl(testfd, VIDIOC_G_STD, &std) < 0)
+        perror("VIDIOC_G_STD");
+    else
     {
         if (std & V4L2_STD_PAL)
             ntsc = false;
@@ -248,49 +282,97 @@
     memcpy(prvpkt, buf + len - 3, 3);
 }
 
-void IvtvDecoder::GetFrame(int onlyvideo)
+bool IvtvDecoder::ReadWrite(int onlyvideo, int delay)
 {
+    if (ateof)
+        return false;
+
     gotvideo = false;
     frame_decoded = 0;
 
-    bool allowedquit = false;
+    int count, total = 0;
 
-    unsigned char buf[256001];
+    if (onlyvideo < 0)
+        vidread = vidwrite = vidfull = 0;
+    else if (vidfull || vidread != vidwrite)
+    {
+        VideoOutputIvtv *videoout = (VideoOutputIvtv *)m_parent->getVideoOutput();
+        if (delay)
+            videoout->Poll(delay);
 
-    VideoOutputIvtv *videoout = (VideoOutputIvtv *)m_parent->getVideoOutput();
-    long long startpos = ringBuffer->GetReadPosition();
-    int count = 0;
-    int newframes = 0;
-    int onlywrote = 0;
-    
-    while (!allowedquit)
+        if (vidwrite >= vidread)
+        {
+            count = videoout->WriteBuffer(&vidbuf[vidwrite], vidmax - vidwrite);
+            if (count < 0)
+                ateof = true;
+            else if (count > 0)
+            {
+                vidwrite = (vidwrite + count) & (vidmax - 1);
+                vidfull = 0;
+                total += count;
+                //fprintf(stderr, "write1 = %d, %d, %d, %d\n", count, vidread, vidwrite, vidfull);
+            }
+        }
+
+        if (vidwrite < vidread)
+        {
+            count = videoout->WriteBuffer(&vidbuf[vidwrite], vidread - vidwrite);
+            if (count < 0)
+                ateof = true;
+            else if (count > 0)
+            {
+                vidwrite = (vidwrite + count) & (vidmax - 1);
+                vidfull = 0;
+                total += count;
+                //fprintf(stderr, "write2 = %d, %d, %d, %d\n", count, vidread, vidwrite, vidfull);
+            }
+        }
+    }
+
+    if (!vidfull || vidread != vidwrite)
     {
-        count = ringBuffer->Read(buf, 131072);
+        if (vidread >= vidwrite)
+        {
+            long long startpos = ringBuffer->GetReadPosition();
+            count = ringBuffer->Read(&vidbuf[vidread], vidmax - vidread);
+            if (count > 0)
+            {
+                MpegPreProcessPkt(&vidbuf[vidread], count, startpos);
+                vidread = (vidread + count) & (vidmax - 1);
+                vidfull = (vidread == vidwrite);
+                total += count;
+                //fprintf(stderr, "read1 = %d, %d, %d, %d\n", count, vidread, vidwrite, vidfull);
+            }
+        }
 
-        MpegPreProcessPkt(buf, count, startpos);
-        if (onlyvideo >= 0)
+        if (vidread < vidwrite)
         {
-            ingetframe = true;
-            onlywrote = videoout->WriteBuffer(buf, count, newframes);
-            ingetframe = false;
-	}
-	    
-        allowedquit = true;
-    }                    
+            long long startpos = ringBuffer->GetReadPosition();
+            count = ringBuffer->Read(&vidbuf[vidread], vidwrite - vidread);
+            if (count > 0)
+            {
+                MpegPreProcessPkt(&vidbuf[vidread], count, startpos);
+                vidread = (vidread + count) & (vidmax - 1);
+                vidfull = (vidread == vidwrite);
+                total += count;
+                //fprintf(stderr, "read2 = %d, %d, %d, %d\n", count, vidread, vidwrite, vidfull);
+            }
+        }
 
-    if (newframes > framesPlayed)
-        framesPlayed = newframes;
+        if (total == 0)
+            ateof = true;
+    }
 
-    m_parent->SetFramesPlayed(framesPlayed);
+    if (ateof)
+        m_parent->SetEof();
+
+    return (total > 0);
 }
 
-void IvtvDecoder::InterruptDisplay(void)
+void IvtvDecoder::GetFrame(int onlyvideo)
 {
-    if (ingetframe)
-    {
-        VideoOutputIvtv *videoout = (VideoOutputIvtv *)m_parent->getVideoOutput();
-        videoout->InterruptDisplay();
-    }
+    ReadWrite(onlyvideo, 1);
+    UpdateFramesPlayed();
 }
 
 bool IvtvDecoder::DoRewind(long long desiredFrame)
@@ -305,7 +387,6 @@
     if (lastKey < 0)
         lastKey = 0;
 
-    int normalframes = desiredFrame - lastKey;
     long long keyPos = positionMap[lastKey / keyframedist];
     long long curPosition = ringBuffer->GetTotalReadPosition();
     long long diff = keyPos - curPosition;
@@ -317,12 +398,10 @@
         if (keyPos == 0)
             continue;
         diff = keyPos - curPosition;
-        normalframes = 0;
         if (lastKey > storelastKey)
         {
             lastKey = storelastKey;
             diff = 0;
-            normalframes = 0;
             return false;
         }
     }
@@ -337,7 +416,6 @@
         lastKey -= keyframedist;
         keyPos = positionMap[lastKey / keyframedist];
         diff = keyPos - curPosition;
-        normalframes += keyframedist;
 
         if (keyPos != 0)
            VERBOSE(VB_PLAYBACK, QString("Using seek position: %1")
@@ -353,31 +431,30 @@
 
         lastKey = storelastKey;
         diff = 0;
-        normalframes = 0;
 
         return false;
     }
 
     ringBuffer->Seek(diff, SEEK_CUR);
 
-    framesPlayed = lastKey;
-    framesRead = lastKey;
-
-    normalframes = desiredFrame - framesPlayed;
+    SeekReset(lastKey, desiredFrame - lastKey);
 
-    if (!exactseeks)
-        normalframes = 0;
-
-    SeekReset(normalframes);
-
-    m_parent->SetFramesPlayed(framesPlayed);
     return true;
 }
 
 bool IvtvDecoder::DoFastForward(long long desiredFrame)
 {
-    lastKey = (framesPlayed / keyframedist) * keyframedist;
     long long number = desiredFrame - framesPlayed;
+
+    if (m_parent->GetPause() && number < 15)
+    {
+        StepFrames(framesPlayed, number+1);
+        framesPlayed = desiredFrame + 1;
+        m_parent->SetFramesPlayed(framesPlayed);
+        return !ateof;
+    }
+
+    lastKey = (framesPlayed / keyframedist) * keyframedist;
     long long desiredKey = lastKey;
 
     while (desiredKey <= desiredFrame)
@@ -386,11 +463,6 @@
     }
     desiredKey -= keyframedist;
    
-    int normalframes = desiredFrame - desiredKey;
-
-    if (desiredKey == lastKey)
-        normalframes = number;
-
     long long keyPos = -1;
 
     long long tmpKey = desiredKey;
@@ -464,16 +536,8 @@
 
     ringBuffer->Seek(diff, SEEK_CUR);
 
-    framesPlayed = framesRead = lastKey;
-
-    normalframes = desiredFrame - framesPlayed;
+    SeekReset(lastKey, desiredFrame - lastKey);
 
-    if (!exactseeks)
-        normalframes = 0;
-
-    SeekReset(normalframes);
-
-    m_parent->SetFramesPlayed(framesPlayed);
     return true;
 }
 
@@ -483,3 +547,76 @@
         m_playbackinfo->SetPositionMap(positionMap, MARK_GOP_START, m_db);
 }
 
+void IvtvDecoder::UpdateFramesPlayed(void)
+{
+    VideoOutputIvtv *videoout = (VideoOutputIvtv *)m_parent->getVideoOutput();
+
+    int newframes = videoout->GetFramesPlayed();
+
+    if (lastStartFrame + newframes >= framesPlayed &&
+        lastStartFrame + newframes <= framesRead)
+        framesPlayed = lastStartFrame + newframes;
+    else
+        fprintf(stderr, "bad newframes: %d %lld %lld %lld\n",
+                newframes, lastStartFrame, framesRead, framesPlayed);
+
+    m_parent->SetFramesPlayed(framesPlayed);
+}
+
+bool IvtvDecoder::StepFrames(int start, int count)
+{
+    VideoOutputIvtv *videoout = (VideoOutputIvtv *)m_parent->getVideoOutput();
+
+    int step, cur = 0, last = start;
+
+    //fprintf(stderr, "stepping %d from %d\n", count, last);
+
+    for (step = 0; step < count && !ateof; step++)
+    {
+        while (ReadWrite(1, 0))
+            ;
+
+        //fprintf(stderr, "    step %d at %d\n", step, last);
+        videoout->Step();
+        usleep(1000);
+
+        int tries;
+        const int maxtries = 500000;
+        const int restep = 100000;
+
+        for (tries = 0; tries < maxtries && !ateof; tries++)
+        {
+            while (ReadWrite(1, 0))
+                ;
+
+            cur = lastStartFrame + videoout->GetFramesPlayed();
+            if (cur > last)
+                break;
+
+            if (tries && !(tries % restep))
+            {
+                videoout->Pause();
+                fprintf(stderr, "        extra step %d at %d\n", step, last);
+                videoout->Step();
+                usleep(1000);
+            }
+        }
+
+        videoout->Pause();
+
+        //fprintf(stderr, "        %d tries to %d\n", tries, cur);
+
+        if (tries >= maxtries)
+        {
+            cerr << "timed out while stepping, giving up" << endl;
+            break;
+        }
+        
+        last = cur;
+    }
+
+    //fprintf(stderr, "    stepped to %d\n", cur);
+
+    return true;
+}
+
Index: libs/libmythtv/ivtvdecoder.h
===================================================================
RCS file: /var/lib/mythcvs/mythtv/libs/libmythtv/ivtvdecoder.h,v
retrieving revision 1.9
diff -u -r1.9 ivtvdecoder.h
--- libs/libmythtv/ivtvdecoder.h	6 Jan 2004 07:21:40 -0000	1.9
+++ libs/libmythtv/ivtvdecoder.h	9 Jan 2004 04:11:56 -0000
@@ -48,21 +48,24 @@
 
     QString GetEncodingType(void) { return QString("MPEG-2"); }
 
-    void InterruptDisplay(void);
+    void UpdateFramesPlayed(void);
 
   protected:
     RingBuffer *getRingBuf(void) { return ringBuffer; }
 
   private:
     void MpegPreProcessPkt(unsigned char *buf, int len, long long startpos);
-    void SeekReset(int skipframes = 0);
+    void SeekReset(long long newkey, int skipframes = 0);
+    bool ReadWrite(int onlyvideo, int delay);
+    bool StepFrames(int start, int count);
 
     RingBuffer *ringBuffer;
 
     int frame_decoded;
 
-    long long framesPlayed;
     long long framesRead;
+    long long framesPlayed;
+    long long lastStartFrame;
 
     bool hasFullPositionMap;
     QMap<long long, long long> positionMap;
@@ -92,7 +95,9 @@
 
     static bool ntsc;
 
-    bool ingetframe;
+    static const int vidmax = 131072; // must be a power of 2
+    unsigned char vidbuf[vidmax];
+    int vidread, vidwrite, vidfull;
 };
 
 #endif
Index: libs/libmythtv/tv_play.cpp
===================================================================
RCS file: /var/lib/mythcvs/mythtv/libs/libmythtv/tv_play.cpp,v
retrieving revision 1.139
diff -u -r1.139 tv_play.cpp
--- libs/libmythtv/tv_play.cpp	3 Jan 2004 02:58:04 -0000	1.139
+++ libs/libmythtv/tv_play.cpp	9 Jan 2004 04:11:57 -0000
@@ -986,7 +979,10 @@
         if (!nvp->DoKeypress(e))
             editmode = nvp->GetEditMode();
         if (!editmode)
-            nvp->SetPlaySpeed(1.0, true);
+        {
+            paused = !paused;
+            DoPause();
+        }
         return;
     }
 
@@ -1074,8 +1070,11 @@
                     {
                        playbackinfo->SetEditing(false, m_db);
                        editmode = nvp->EnableEdit();
-                       if (editmode)
-                           nvp->SetPlaySpeed(1.0, false);
+                    }
+                    else
+                    {
+                        paused = !paused;
+                        DoPause();
                     }
                 }
                 else if (dialogname == "exitplayoptions") 
@@ -1091,7 +1090,8 @@
                             exitPlayer = true;
                             break;
                         case 3: case 0:
-                            nvp->Unpause();
+                            paused = !paused;
+                            DoPause();
                             break;
                         case 4:
                             wantsToQuit = true;
@@ -1226,8 +1226,6 @@
             DoSeek(jumptime * 60, tr("Jump Ahead"));
         else if (action == "ESCAPE")
         {
-            StopFFRew();
-
             if (osd)
             {
                 QStringList osetname;
@@ -1239,6 +1237,9 @@
                 }
             }
 
+            NormalSpeed();
+            StopFFRew();
+
             if (StateIsPlaying(internalState) &&
                 gContext->GetNumSetting("PlaybackExitPrompt") == 1) 
             {
@@ -1297,7 +1298,7 @@
             {
                 float time = StopFFRew();
                 UpdatePosOSD(time, tr("Play"));
-                was_doing_ff_rew = true;
+                handled = true;
             }
         }
 
@@ -1305,7 +1306,7 @@
         {
             NormalSpeed();
             UpdatePosOSD(0.0, tr("Play"));
-            was_doing_ff_rew = true;
+            handled = true;
         }
     }
 
@@ -1408,8 +1409,6 @@
                     return;
                 }
                 editmode = nvp->EnableEdit();
-                if (editmode)
-                    nvp->SetPlaySpeed(1.0, false);
             }
             else if (action == "TOGGLEBROWSE")
                 DoProgramMenu();
@@ -1529,10 +1530,15 @@
 
 void TV::DoPause(void)
 {
-    NormalSpeed();
+    speed_index = 0;
     float time = StopFFRew();
 
-    paused = activenvp->TogglePause();
+    if (paused)
+        activenvp->Play(1.0, true);
+    else
+        activenvp->Pause();
+
+    paused = !paused;
 
     if (activenvp != nvp)
         return;
@@ -1540,7 +1546,7 @@
     if (paused)
         UpdatePosOSD(time, tr("Paused"));
     else
-        osd->EndPause();
+        UpdatePosOSD(time, tr("Play"));
 }
 
 void TV::DoInfo(void)
@@ -1590,7 +1596,7 @@
         QString desc = "";
         int pos = nvp->calcSliderPos(desc);
         bool slidertype = (internalState == kState_WatchingLiveTV);
-        int disptime = (mesg == tr("Paused")) ? -1 : 2;
+        int disptime = (mesg == tr("Paused")) ? -1 : 2;
         int osdtype = (doSmartForward) ? kOSDFunctionalType_SmartForward :
                                          kOSDFunctionalType_Default;
 
@@ -1633,7 +1639,7 @@
         return;
 
     speed_index = 0;
-    activenvp->SetPlaySpeed(1.0, true);
+    activenvp->Play(1.0, true);
 }
 
 void TV::ChangeSpeed(int direction)
@@ -1662,11 +1668,14 @@
         default: speed_index = old_speed; return; break;
     }
 
-    activenvp->SetPlaySpeed(speed, (speed == 1.0));
-    UpdatePosOSD(time, mesg);
+    if (!activenvp->Play(speed, (speed == 1.0)))
+    {
+        speed_index = old_speed;
+        return;
+    }
 
-    if (paused)
-        paused = activenvp->TogglePause();
+    paused = false;
+    UpdatePosOSD(time, mesg);
 }
 
 float TV::StopFFRew(void)
@@ -1687,7 +1696,7 @@
     doing_ff_rew = 0;
     ff_rew_index = SSPEED_NORMAL;
 
-    activenvp->SetPlaySpeed(1.0, true);
+    activenvp->Play(1.0, true);
 
     return time;
 }
@@ -1713,10 +1722,8 @@
         doing_ff_rew = direction;
         ff_rew_index = SSPEED_NORMAL;
 
-        if (paused)
-            paused = activenvp->TogglePause();
-
-        activenvp->SetPlaySpeed(1.0, false);
+        activenvp->Play(1.0, false);
+        paused = false;
     }
 }
 
@@ -1837,7 +1844,7 @@
     activerecorder->ToggleInputs();
 
     activenvp->ResetPlaying();
-    activenvp->Unpause(false);
+    activenvp->Play(1.0, true, false);
 
     if (activenvp == nvp)
         UpdateOSDInput();
@@ -1878,7 +1885,7 @@
     activerecorder->ChangeChannel(direction);
 
     activenvp->ResetPlaying();
-    activenvp->Unpause(false);
+    activenvp->Play(1.0, true, false);
 
     if (activenvp == nvp)
     {
@@ -2030,7 +2037,7 @@
     activerecorder->SetChannel(name);
 
     activenvp->ResetPlaying();
-    activenvp->Unpause(false);
+    activenvp->Play(1.0, true, false);
 
     if (activenvp == nvp)
     {
Index: libs/libmythtv/videoout_ivtv.cpp
===================================================================
RCS file: /var/lib/mythcvs/mythtv/libs/libmythtv/videoout_ivtv.cpp,v
retrieving revision 1.18
diff -u -r1.18 videoout_ivtv.cpp
--- libs/libmythtv/videoout_ivtv.cpp	10 Nov 2003 21:45:01 -0000	1.18
+++ libs/libmythtv/videoout_ivtv.cpp	9 Jan 2004 04:11:57 -0000
@@ -37,23 +37,13 @@
     pixels = NULL;
     lastcleared = false;
     videoDevice = "/dev/video16";
-    skipplay = false;
-    interruptdisplay = false;
+    last_speed = 1.0;
+    last_normal = true;
 }
 
 VideoOutputIvtv::~VideoOutputIvtv()
 {
-    if (videofd)
-    {
-        ivtv_cfg_stop_decode sd;
-        memset(&sd, 0, sizeof(sd));
-
-        sd.hide_last = 1;
-        ioctl(videofd, IVTV_IOC_S_STOP_DECODE, &sd);
-
-        close(videofd);
-    }
-    videofd = -1;
+    Close();
 
     if (fbfd > 0)
     {
@@ -91,7 +81,8 @@
         struct ivtv_osd_coords osdcoords;
         memset(&osdcoords, 0, sizeof(osdcoords));
 
-        ioctl(fbfd, IVTVFB_IOCTL_GET_ACTIVE_BUFFER, &osdcoords);
+        if (ioctl(fbfd, IVTVFB_IOCTL_GET_ACTIVE_BUFFER, &osdcoords) < 0)
+            perror("IVTVFB_IOCTL_GET_ACTIVE_BUFFER");
 
         struct ivtvfb_ioctl_dma_host_to_ivtv_args prep;
         memset(&prep, 0, sizeof(prep));
@@ -102,7 +93,8 @@
 
         memset(osdbuf_aligned, 0x00, osdbufsize);
 
-        ioctl(fbfd, IVTVFB_IOCTL_PREP_FRAME, &prep);
+        if (ioctl(fbfd, IVTVFB_IOCTL_PREP_FRAME, &prep) < 0)
+            perror("IVTVFB_IOCTL_PREP_FRAME");
 
         usleep(20000);
 
@@ -110,7 +102,8 @@
         osdcoords.offset = 0;
         osdcoords.pixel_stride = XJ_width * 2;
 
-        ioctl(fbfd, IVTVFB_IOCTL_SET_ACTIVE_BUFFER, &osdcoords);
+        if (ioctl(fbfd, IVTVFB_IOCTL_SET_ACTIVE_BUFFER, &osdcoords) < 0)
+            perror("IVTVFB_IOCTL_SET_ACTIVE_BUFFER");
     }
 }
 
@@ -138,7 +131,7 @@
 
     MoveResize();
 
-    Reopen();
+    Open();
 
     if (videofd <= 0)
         return false;
@@ -195,7 +188,8 @@
         struct ivtvfb_ioctl_get_frame_buffer igfb;
         memset(&igfb, 0, sizeof(igfb));
 
-        ioctl(fbfd, IVTVFB_IOCTL_GET_FRAME_BUFFER, &igfb);
+        if (ioctl(fbfd, IVTVFB_IOCTL_GET_FRAME_BUFFER, &igfb) < 0)
+            perror("IVTVFB_IOCTL_GET_FRAME_BUFFER");
 
         stride = igfb.sizex * 4;
 
@@ -212,62 +206,41 @@
     return true;
 }
 
-void VideoOutputIvtv::Reopen(int skipframes, int newstartframe)
+void VideoOutputIvtv::Close(void)
 {
-    if (videofd >= 0)
-    {
-        ivtv_cfg_stop_decode sd;
-        memset(&sd, 0, sizeof(sd));
-
-        ioctl(videofd, IVTV_IOC_S_STOP_DECODE, &sd);
-        close(videofd);
-
-        skipplay = true;
-    }
+    if (videofd < 0)
+        return;
 
+    close(videofd);
     videofd = -1;
+}
+
+void VideoOutputIvtv::Open(void)
+{
+    if (videofd >= 0)
+        return;
 
-    if ((videofd = open(videoDevice.ascii(), O_WRONLY | O_LARGEFILE, 0555)) < 0)
+    if ((videofd = open(videoDevice.ascii(), 
+        O_WRONLY | O_LARGEFILE | O_NONBLOCK, 0555)) < 0)
     {
         perror("Cannot open ivtv video out device");
         return;
     }
-    else
-    {
-        struct v4l2_control ctrl;
-        memset(&ctrl, 0, sizeof(ctrl));
 
-        ctrl.id = V4L2_CID_IVTV_DEC_PREBUFFER;
-        ctrl.value = 0;
+    struct v4l2_control ctrl;
+    memset(&ctrl, 0, sizeof(ctrl));
 
-        ioctl(videofd, VIDIOC_S_CTRL, &ctrl);
+    ctrl.id = V4L2_CID_IVTV_DEC_PREBUFFER;
+    ctrl.value = 1;
 
-        ctrl.id = V4L2_CID_IVTV_DEC_NUM_BUFFERS;
-        ctrl.value = 0;
+    if (ioctl(videofd, VIDIOC_S_CTRL, &ctrl) < 0)
+        perror("VIDIOC_S_CTRL prebuffer");
 
-        ioctl(videofd, VIDIOC_S_CTRL, &ctrl);
-        
-        ivtv_cfg_start_decode startd;
-        memset(&startd, 0, sizeof(startd));
+    ctrl.id = V4L2_CID_IVTV_DEC_NUM_BUFFERS;
+    ctrl.value = 1;
 
-        if (skipframes > 0)
-        {
-            startd.gop_offset = skipframes;
-            startd.muted_audio_frames = 6;
-        }
-
-        ioctl(videofd, IVTV_IOC_S_START_DECODE, &startd);
-
-        // change it back to hide last frame, in case we die unexpectedly.
-        ivtv_cfg_stop_decode stopd;
-        memset(&stopd, 0, sizeof(stopd));
-
-        stopd.hide_last = 1;
-        ioctl(videofd, IVTV_IOC_S_STOP_DECODE, &stopd);
-    }
-
-    startframenum = newstartframe;
-    firstframe = true;
+    if (ioctl(videofd, VIDIOC_S_CTRL, &ctrl) < 0)
+        perror("VIDIOC_S_CTRL numbuffers");
 }
 
 void VideoOutputIvtv::EmbedInWidget(unsigned long wid, int x, int y, int w, 
@@ -366,89 +339,155 @@
             prep.dest_offset = 0;
             prep.count = XJ_height * stride;
 
-            ioctl(fbfd, IVTVFB_IOCTL_PREP_FRAME, &prep);
+            if (ioctl(fbfd, IVTVFB_IOCTL_PREP_FRAME, &prep) < 0)
+                perror("IVTVFB_IOCTL_PREP_FRAME");
         }
     }
 }
 
-void VideoOutputIvtv::Play()
+void VideoOutputIvtv::Start(int skip, int mute)
 {
-    if (!skipplay)
-        ioctl(videofd, IVTV_IOC_PLAY, 0);
-    skipplay = false;
+    struct ivtv_cfg_start_decode start;
+    memset(&start, 0, sizeof start);
+    start.gop_offset = skip;
+    start.muted_audio_frames = mute;
+
+    while (ioctl(videofd, IVTV_IOC_START_DECODE, &start) < 0)
+    {
+        if (errno != EBUSY)
+        {
+            perror("IVTV_IOC_START_DECODE");
+            break;
+        }
+    }
 }
 
-void VideoOutputIvtv::Pause()
+void VideoOutputIvtv::Stop(void)
 {
-    ioctl(videofd, IVTV_IOC_PAUSE, 0);
+    struct ivtv_cfg_stop_decode stop;
+    memset(&stop, 0, sizeof stop);
+    stop.hide_last = 0;
+
+    while (ioctl(videofd, IVTV_IOC_STOP_DECODE, &stop) < 0)
+    {
+        if (errno != EBUSY)
+        {
+            perror("IVTV_IOC_STOP_DECODE");
+            break;
+        }
+    }
 }
 
-void VideoOutputIvtv::InterruptDisplay(void)
+void VideoOutputIvtv::Pause(void)
 {
-    interruptdisplay = true;
+    while (ioctl(videofd, IVTV_IOC_PAUSE, 0) < 0)
+    {
+        if (errno != EBUSY)
+        {
+            perror("IVTV_IOC_PAUSE");
+            break;
+        }
+    }
 }
 
-int VideoOutputIvtv::WriteBuffer(unsigned char *buf, int count, int &frames)
+void VideoOutputIvtv::Poll(int delay)
 {
-    int n = 0;
-
     struct pollfd polls;
     polls.fd = videofd;
     polls.events = POLLOUT;
     polls.revents = 0;
 
-    int ret = 0;
-    int totalpassed = count;
-    const int maxwrite = 32768;
-    
-    while (count > 0 && !interruptdisplay)
-    {
-        ret = poll(&polls, 1, 20);
-
-        if (ret == 1 && polls.revents & POLLOUT)
-        {
-            n = write(videofd, buf, (count > maxwrite) ? maxwrite : count);
-            if (n < 0)
-            {
-                perror("Writing to videodev");
-                return n;
-            }
-            count -= n;
-            buf += n;
-        }
+    if (poll(&polls, 1, delay) < 0)
+        perror("Polling on videodev");
+}
+
+int VideoOutputIvtv::WriteBuffer(unsigned char *buf, int len)
+{
+    int count;
 
-        if (interruptdisplay)
+    //cerr << "ivtv writing video... ";
+    count = write(videofd, buf, len);
+    if (count < 0)
+    {
+        if (errno != EAGAIN)
         {
-            break;
+            perror("Writing to videodev");
+            return count;
         }
+        count = 0;
     }
+    //cerr << "wrote " << count << " of " << len << endl;
 
-    interruptdisplay = false;
-       
+    return count;
+}
+
+int VideoOutputIvtv::GetFramesPlayed(void)
+{
     struct ivtv_ioctl_framesync frameinfo;
-    memset(&frameinfo, 0, sizeof(frameinfo));
+    memset(&frameinfo, 0, sizeof frameinfo);
+
     if (ioctl(videofd, IVTV_IOC_GET_TIMING, &frameinfo) < 0)
+        perror("IVTV_IOC_GET_TIMING");
+
+    return frameinfo.frame;
+}
+
+bool VideoOutputIvtv::Play(float speed, bool normal)
+{
+    if (speed > 3.0)
+        return false;
+
+    struct ivtv_speed play;
+    memset(&play, 0, sizeof play);
+    if (speed >= 2.0)
+        play.scale = (int)roundf(speed);
+    else if (speed <= 0.5)
+        play.scale = (int)roundf(1 / speed);
+    else
+        play.scale = 1;
+    play.smooth = 0;
+    play.speed = (speed > 1.0);
+    play.direction = 0;
+    play.fr_mask = 2;
+    play.b_per_gop = 0;
+    play.aud_mute = !normal;
+    play.fr_field = 0;
+    play.mute = 0;
+
+    while (ioctl(videofd, IVTV_IOC_S_SPEED, &play) < 0)
     {
-        perror("IVTV_IOC_FRAMESYNC");
-        frames = 0;
-        return (totalpassed - count);
+        if (errno != EBUSY)
+        {
+            perror("IVTV_IOC_S_SPEED");
+            break;
+        }
     }
 
-    // seems the decoder doesn't initialize this properly.
-    if (frameinfo.frame >= 20000000)
-        frameinfo.frame = 0;
+    last_speed = speed;
+    last_normal = normal;
 
-    if (firstframe)
+    return true;
+}
+
+void VideoOutputIvtv::Flush(void)
+{
+    int arg = 0;
+
+    if (ioctl(videofd, IVTV_IOC_DEC_FLUSH, &arg) < 0)
+        perror("IVTV_IOC_DEC_FLUSH");
+}
+
+void VideoOutputIvtv::Step(void)
+{
+    int arg = 0;
+
+    while (ioctl(videofd, IVTV_IOC_DEC_STEP, &arg) < 0)
     {
-        if (frameinfo.frame > 5)
-            frameinfo.frame = 0;
-        else
-            firstframe = false;
+        if (errno != EBUSY)
+        {
+            perror("IVTV_IOC_DEC_STEP");
+            break;
+        }
     }
-
-    interruptdisplay = false;
-    
-    frames = frameinfo.frame + startframenum;
-    return totalpassed - count;
 }
 
Index: libs/libmythtv/videoout_ivtv.h
===================================================================
RCS file: /var/lib/mythcvs/mythtv/libs/libmythtv/videoout_ivtv.h,v
retrieving revision 1.13
diff -u -r1.13 videoout_ivtv.h
--- libs/libmythtv/videoout_ivtv.h	8 Nov 2003 23:59:26 -0000	1.13
+++ libs/libmythtv/videoout_ivtv.h	9 Jan 2004 04:11:57 -0000
@@ -33,17 +33,24 @@
                       FilterChain *filterList,
                       NuppelVideoPlayer *pipPlayer);
 
-    int WriteBuffer(unsigned char *buf, int count, int &frames);
+    int WriteBuffer(unsigned char *buf, int count);
+    void Poll(int delay);
     void Pause(void);
-    void Play(void);
+    void Start(int skip, int mute);
+    void Stop(void);
 
-    void Reopen(int skipframes = 0, int newstartframe = 0);
+    void Open(void);
+    void Close(void);
 
     void SetFPS(float lfps) { fps = lfps; }
 
     void ClearOSD(void);
 
-    void InterruptDisplay(void);
+    bool Play(float speed, bool normal);
+    bool Play(void) { return Play(last_speed, last_normal); };
+    void Flush(void);
+    void Step(void);
+    int GetFramesPlayed(void);
 
   private:
     int videofd;
@@ -57,8 +64,6 @@
 
     QMutex lock;
 
-    long long startframenum;
-
     int mapped_offset;
     int mapped_memlen;
     char *mapped_mem;
@@ -71,12 +76,10 @@
     char *osdbuffer;
     char *osdbuf_aligned;
 
-    bool firstframe;
-
     int osdbufsize;
 
-    bool skipplay;
-    bool interruptdisplay;
+    float last_speed;
+    bool last_normal;
 };
 
 #endif
